/*
    flame - cosmic recursive fractal flames
    Copyright (C) 1992  Scott Draves <spot@cs.cmu.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "screenhack.h"
#include "libifs.h"


char *progclass = "Flame";

char *defaults [] = {
  "*background:	slate gray",
  "*foreground:	yellow",
  "*length1:	2000000",
  "*length2:	20000000",
  "*batch:	20000",
  0
};

XrmOptionDescRec options [] = {
  { "-iterations",	".iterations",	XrmoptionSepArg, 0 },
  { "-batch",		".batch",	XrmoptionSepArg, 0 },
  { "-length1",		".length1",	XrmoptionSepArg, 0 },
  { "-length2",		".length2",	XrmoptionSepArg, 0 },
};
int options_size = (sizeof (options) / sizeof (options[0]));


/*

basic action:

there's a current control point.  usually it sits and iterates and dumps on
the screen.  every N points, it clears the screen, gets a new random
control point, and continues.  various kbd commands modify the control
point (including reading from a file) (including moving back in the stack
of previous control points), clear the screen, and reset the counter to
something higher than N.  a command exits to append the current control
point to a file (stdout?).  mono.  the kbd is checked every batch-size
points.  there are commands to disable clearing the screen.

editing commands:

*1) append to file
*2) load next from file
*3) add the inverse of the current delta, generate a new random delta, and add
it in
*4) add the current delta again
5) undo stack
6) registers
*6.5) new random control point
7) interpolate between current and register
8) switch all xforms to a variation
9) set current xform
10) set current xform variation
11) increase/decrease all coefs by x%
12) pre/post rotate left/right
13) translate up/down/left/right


a) widen/narrow either eigenvalue
b) hook above function up to mouse value

*/

void
add_delta(delta, cp, mult)
   double delta[NXFORMS][3][2];
   control_point *cp;
   double mult;
{
   int i, j, k;
   for (i = 0; i < NXFORMS; i++)
      for (j = 0; j < 3; j++)
	 for (k = 0; k < 2; k++)
	    cp->xform[i].c[j][k] += mult * delta[i][j][k];
}

#define state_neutral 0

void
screenhack (dpy, window)
     Display *dpy;
     Window window;
{
   GC draw_gc, erase_gc;
   unsigned int default_fg_pixel;

   int batch_size;
   int normal_length;
   int special_length;
   int main_counter;

   int width;
   int height;
   
   control_point cp;
   double delta[NXFORMS][3][2];
   double delta_size = 0.10;
   point *points;
   XPoint *xpts;
#define inc_prev_count \
{prev_count++; \
 if (prev_count == NFRAMES) prev_count = 0; prevs[prev_count] = cp;\
    prevs[prev_count].time = back_time = prev_count;}

   int done = 0, prev_count = 0;
#define NFRAMES 1000
   control_point *frames = (control_point *)
      malloc(sizeof(control_point) * NFRAMES);
   control_point *prevs = (control_point *)
      malloc(sizeof(control_point) * NFRAMES);
   int param = 0;
   double cur_time = 0.0, dtime = 0.2, dtheta = M_PI / 2.0;
   double back_time = 0.0;
   char sbuffer[1000];

   if (1) {
      Colormap cmap;
      XWindowAttributes xgwa;
      XSetWindowAttributes xswa;
      XGCValues gcv;
      
      batch_size = get_integer_resource ("batch", "Integer");
      normal_length = get_integer_resource ("length1", "Integer");
      special_length = get_integer_resource ("length2", "Integer");
      
      xswa.event_mask = KeyPressMask;
      XChangeWindowAttributes (dpy, window, CWEventMask, &xswa);
      
      XGetWindowAttributes (dpy, window, &xgwa);
      cmap = xgwa.colormap;
      gcv.foreground = default_fg_pixel =
	 get_pixel_resource ("foreground", "Foreground", dpy, cmap);
      draw_gc = XCreateGC (dpy, window, GCForeground, &gcv);
      gcv.foreground = get_pixel_resource ("background", "Background", dpy, cmap);
      erase_gc = XCreateGC (dpy, window, GCForeground, &gcv);
      XClearWindow (dpy, window);
      XSetForeground (dpy, draw_gc, default_fg_pixel);
      width = xgwa.width;
      height = xgwa.height;
      
      
      
      main_counter = normal_length;
      random_control_point(&cp, variation_random);
      prevs[prev_count] = cp;
      get_cmap(cmap_random, cp.cmap, 256);
      if (1) {
	 int i, j, k;
	 for (i = 0; i < NXFORMS; i++)
	    for (j = 0; j < 3; j++)
	       for (k = 0; k < 2; k++)
		  delta[i][j][k] = 0.0;
	 for (i = 0; i < NFRAMES; i++) {
	    frames[i].time = (double) i;
	    prevs[i].time = (double) i;
	 }
      }
      points = (point *) malloc(sizeof(point) * batch_size);
      xpts = (XPoint *) malloc(sizeof(XPoint) * batch_size);
      
   }
   
   while (!done) {
      int i;
      XEvent xev;
      int state = state_neutral;
      points[0][0] = random_uniform11();
      points[0][1] = random_uniform11();
      points[0][2] = 0.0;
      
      iterate(&cp, batch_size, 10, points);
      for (i = 0; i < batch_size; i++) {
	 xpts[i].x = (points[i][0] + 1.0) * width/2;
	 xpts[i].y = (points[i][1] + 1.0) * height/2;
      }
      
      XDrawPoints(dpy, window, draw_gc, xpts, batch_size, CoordModeOrigin);

      if (main_counter > 0) {
	 main_counter -= batch_size;
	 if (main_counter <= 0) {
	    random_control_point(&cp, variation_random);
	    XClearWindow (dpy, window);
	    inc_prev_count;
	    main_counter = normal_length;
	 }
      }

      while (XCheckWindowEvent(dpy, window, -1, &xev)) {
	 char user_string[20];
	 int user_string_index, len;
	 switch (xev.type) {
	  case KeyPress:
	    len = XLookupString((XKeyEvent *) &xev, user_string, 20, 0, 0);
	    for (user_string_index = 0;
		 user_string_index < len; user_string_index++) {
	       char this_char = user_string[user_string_index];
	       switch (state) {
		case state_neutral:
		  /* printf("%c\n", this_char); */
		  switch (this_char) {
		   case 'q':
		   case 'Q':
		     exit(0);
		   case 'c':
		   case 'C':
		     XClearWindow (dpy, window);
		     break;
		   case 'r':
		     random_control_point(&cp, variation_random);
		     XClearWindow (dpy, window);
		     main_counter = normal_length;
		     inc_prev_count;
		     break;
		   case 's':
		     print_control_point(stdout, &cp, 0);
		     break;
		   case 'f':
		     if (1) {
			char *ss = sbuffer;
			int c, i = 0;
			while (1) {
			   sbuffer[i++] = c = getchar();
			   if (EOF == c)
			      break;
			   if (';' == c) {
			      parse_control_point(&ss, &cp);
			      break;
			   }
			}
		     }
		     inc_prev_count;
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     break;
		   case 'o':
		     if (main_counter > 0)
			main_counter = -1;
		     else
			main_counter = normal_length;
		     break;
		   case 'd':
		     if (1) {
			double s, t = 0.0;
			int i, j, k;
			add_delta(delta, &cp, -delta_size);
			for (i = 0; i < NXFORMS; i++)
			   for (j = 0; j < 3; j++)
			      for (k = 0; k < 2; k++) {
				 s = delta[i][j][k] = random_uniform11();
				 t += s * s;
			      }
			t = 1.0 / sqrt(t);
			for (i = 0; i < NXFORMS; i++)
			   for (j = 0; j < 3; j++)
			      for (k = 0; k < 2; k++)
				 delta[i][j][k] *= t;
			add_delta(delta, &cp, delta_size);
		     }
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case ' ':
		     add_delta(delta, &cp, delta_size);
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case 'e':
		     add_delta(delta, &cp, -delta_size);
		     delta_size = delta_size / 2.0;
		     add_delta(delta, &cp, delta_size);
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case 'E':
		     add_delta(delta, &cp, -delta_size);
		     delta_size = delta_size * 2.0;
		     add_delta(delta, &cp, delta_size);
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case '0':
		   case '1':
		   case '2':
		   case '3':
		   case '4':
		   case '5':
		   case '6':
		   case '7':
		   case '8':
		   case '9':
		     param = 10 * param + (this_char - '0');
		     break;
		   case 'p':
		   case 'P':
		     if (param >= 0 && param < NFRAMES) {
			frames[param] = cp;
			frames[param].time = param;
		     }
		     cur_time = (double) param;
		     param = 0;
		     break;
		   case 'i':
		   case 'I':
		     if (param >= 0 && param < NFRAMES) {
			int i;
			for (i = NFRAMES-1; i > param ; i--) {
			   frames[i] = frames[i-1];
			   frames[i].time = i;
			}
			frames[param] = cp;
			frames[param].time = param;
		     }
		     cur_time = (double) param;
		     param = 0;
		     break;
		   case 'x':
		     if (param >= 0 && param < NFRAMES) {
			int i;
			for (i = param; i > NFRAMES-1; i++) {
			   frames[i] = frames[i+1];
			   frames[i].time = i;
			}
		     }
		     cur_time = (double) param;
		     param = 0;
		     break;
		   case 'g':
		   case 'G':
		     if (param >= 0 && param < NFRAMES)
			cp = frames[param];
		     cur_time = (double) param;
		     param = 0;
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case 'F':
		     if (1) {
			char *ss = sbuffer;
			int c, i = 0, n = param;
			while (1) {
			   i = 0;
			   ss = sbuffer;
			   do {
			      c = getchar();
			      if (EOF == c)
				 goto done_reading;
			      sbuffer[i++] = c;
			   } while (';' != c);
			   get_cmap(cmap_random, frames[n].cmap, 256);
			   parse_control_point(&ss, frames + n);
			   frames[n].time = (double) n;
			   n++;
			}
			done_reading:;
		     }
		     break;
		   case 'n':
		     if (param > 0)
			dtime = 1.0 / param;
		     else
			dtime = 0.2;
		     param = 0;
		     break;
		   case 'N':
		     if (param > 0)
			dtheta = 2.0 * M_PI / param;
		     else
			dtheta = M_PI / 2.0;
		     param = 0;
		     break;
		   case '.':
		     cur_time += 2 * dtime;
		   case ',':
		     cur_time -= dtime;
		     interpolate(frames, NFRAMES, cur_time, &cp);
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     break;
		   case 'v':
		     if (1) {
			int i;
			for (i = 0; i < NXFORMS; i++)
			   if (cp.xform[i].density > 0)
			      stat_matrix(stdout, cp.xform[i].c);
		     }
		     break;
		   case 't':
		     printf("time = %g\n", cur_time);
		     break;
		   case 'y':
		     /* rotate the linear part of all xforms by pi/2 */
		     if (1) {
			int i;
			double s, t;
			for (i = 0; i < NXFORMS; i++) {
			   s = cp.xform[i].c[0][0];
			   t = cp.xform[i].c[0][1];
			   cp.xform[i].c[0][0] = -cp.xform[i].c[1][0];
			   cp.xform[i].c[0][1] = -cp.xform[i].c[1][1];
			   cp.xform[i].c[1][0] = s;
			   cp.xform[i].c[1][1] = t;
			}
		     }
		     XClearWindow (dpy, window);
		     inc_prev_count;
		     main_counter = special_length;
		     break;
		   case 'Y':
		     /* rotate the linear part of all xforms by 2pi/n */
		     if (1) {
			int i;
			double s, t;
			for (i = 0; i < NXFORMS; i++) {
			   double R[2][2];
			   double T[2][2];
			   double U[2][2];
			   R[1][1] = R[0][0] = cos(dtheta);
			   R[0][1] = sin(dtheta);
			   R[1][0] = -R[0][1];
			   T[0][0] = cp.xform[i].c[0][0];
			   T[1][0] = cp.xform[i].c[1][0];
			   T[0][1] = cp.xform[i].c[0][1];
			   T[1][1] = cp.xform[i].c[1][1];
			   mult_matrix(R, T, U);
			   cp.xform[i].c[0][0] = U[0][0];
			   cp.xform[i].c[1][0] = U[1][0];
			   cp.xform[i].c[0][1] = U[0][1];
			   cp.xform[i].c[1][1] = U[1][1];
			}
		     }
		     XClearWindow (dpy, window);
		     inc_prev_count;
		     main_counter = special_length;
		     break;
		   case 'w':
		     /* rotates the type fields of all the xforms
			rotates to the LEFT */
		     if (1) {
			int i, j;
			double t;
			for (i = 0; i < NXFORMS; i++) {
			   t = cp.xform[i].var[0];
			   for (j = 0; j < NVARS-1; j++)
			      cp.xform[i].var[j] = cp.xform[i].var[j+1];
			   cp.xform[i].var[NVARS-1] = t;
			}
		     }
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case 'W':
		     /* rotates the type fields of all the xforms
			rotates to the RIGHT */
		     if (1) {
			int i, j;
			double t;
			for (i = 0; i < NXFORMS; i++) {
			   t = cp.xform[i].var[NVARS-1];
			   for (j = NVARS-1; j > 0; j--)
			      cp.xform[i].var[j] = cp.xform[i].var[j-1];
			   cp.xform[i].var[0] = t;
			}
		     }
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case 'u':
		     /* undo */
		     prev_count--;
		     if (prev_count < 0)
			prev_count = NFRAMES-1;
		     cp = prevs[prev_count];
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     break;
		   case 'U':
		     /* redo */
		     prev_count++;
		     if (prev_count == NFRAMES)
			prev_count = 0;
		     cp = prevs[prev_count];
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     break;
		   case 'a':
		     /* interpolate back into time */
		     back_time += -2.0 * dtime;
		   case 'A':
		     /* and forward */
		     back_time += dtime;
		     interpolate(prevs, NFRAMES, back_time, &cp);
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     printf("back_time = %g\n", back_time);
		     break;
		   case 'h':
		     /* shrink */
		     if (1) {
			int i, j;
			for (i = 0; i < NXFORMS; i++)
			   for (j = 0; j < 3; j++) {
			      cp.xform[i].c[j][0] *= 0.95;
			      cp.xform[i].c[j][1] *= 0.95;
			   }
		     }
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case 'H':
		     /* grow */
		     if (1) {
			int i, j;
			for (i = 0; i < NXFORMS; i++)
			   for (j = 0; j < 3; j++) {
			      cp.xform[i].c[j][0] *= 1.05;
			      cp.xform[i].c[j][1] *= 1.05;
			   }
		     }
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		   case 'S':
		     if (1) {
			int i;
			for (i = 0; i < NXFORMS; i++)
			   cp.xform[i].c[2][1] += -0.1;
		     }
		     goto xlate;
		   case 'X':
		     if (1) {
			int i;
			for (i = 0; i < NXFORMS; i++)
			   cp.xform[i].c[2][1] += 0.1;
		     }
		     goto xlate;
		   case 'K':
		     if (1) {
			int i;
			for (i = 0; i < NXFORMS; i++)
			   cp.xform[i].c[2][0] += -0.1;
		     }
		     goto xlate;
		   case 'L':
		     if (1) {
			int i;
			for (i = 0; i < NXFORMS; i++)
			   cp.xform[i].c[2][0] += 0.1;
		     }
		   xlate:
		     XClearWindow (dpy, window);
		     main_counter = special_length;
		     inc_prev_count;
		     break;
		  }
	       }
	    }
	    break;
	 }
      }
   }
}

